"""
Contains the mesh class and helper classes to work on and generate meshes.
"""
from dataclasses import dataclass
from struct import pack, unpack
import numpy as np
import gmsh
from . import utils
from . import geometry as geo
from .simtwin_base_classes import SimtwinEnum, SimtwinObjectBase


class MeshType(SimtwinEnum):
    """
    Supported types of mesh
    """
    LINEAR = 1
    SERENDIPITY = 2
    FULL = 3


class MeshingMethod(SimtwinEnum):
    """
    Supported meshing methods
    """
    TET = 1
    HEX = 2


class Mesh(SimtwinObjectBase):
    """
    Represents a mesh
    """

    # pylint: disable=R0902
    def __init__(self, server, obj_index=None, geometry=None):
        self._msh_type = MeshType.LINEAR
        self._method = MeshingMethod.TET
        self._sizing = -1.
        self._nc = np.zeros([3, 0])
        self._inz = {}
        self._gp_loc = {}
        self._sj = {}
        self._srf_nodes = np.zeros(0, dtype=int)
        self._srf_inz = {}
        if isinstance(geometry, geo.Geometry):
            self._geometry = geometry
        elif isinstance(obj_index, int):
            self._geometry = None
        else:
            self._geometry = geo.Geometry(server)
        self.methods = {'generate': self.generate}
        super().__init__(server, obj_index)

    def __eq__(self, other):
        if self.refresh_stamp == -1 or other.refresh_stamp == -1:
            return self.obj_index == other.obj_index
        self.__lazy_pull__()
        res = True
        if not self.msh_type == other.msh_type and \
                self.method == other.method and \
                self.nc.shape == other.nc.shape:
            res = False
        if not (self.nc == other.nc).all():
            res = False
        if not utils.cmp_dict(self.inz, other.inz):
            res = False

        if not utils.cmp_dict(self.srf_inz, other.srf_inz):
            res = False

        for et, inz in self.inz.items():
            if not (inz.shape == other.inz[et].shape and (inz == other.inz[et]).all()):
                res = False

        for et, inz in self.srf_inz.items():
            if not (inz.shape == other.srf_inz[et].shape and (inz == other.srf_inz[et]).all()):
                res = False
        return res

    def __hash__(self):
        self.__lazy_pull__()
        return self.obj_index

    def __serialize__(self):
        res = bytes()
        res += self.obj_index.to_bytes(8, 'little')
        data = self._msh_type.__serialize__()
        res += data

        data = self.method.__serialize__()
        res += data

        res += pack('<d', self.sizing)

        data = utils.obj_to_bytes(self._nc)
        res += len(data).to_bytes(8, 'little')
        res += data

        data = utils.serialize_dict(self._inz)
        res += len(data).to_bytes(8, 'little')
        res += data

        data = utils.serialize_dict(self._gp_loc)
        res += len(data).to_bytes(8, 'little')
        res += data

        data = utils.serialize_dict(self._sj)
        res += len(data).to_bytes(8, 'little')
        res += data

        data = utils.obj_to_bytes(self._srf_nodes)
        res += len(data).to_bytes(8, 'little')
        res += data

        data = utils.serialize_dict(self._srf_inz)
        res += len(data).to_bytes(8, 'little')
        res += data

        res += self.geometry.obj_index.to_bytes(8, 'little')
        return res

    def __deserialize__(self, data):
        seeker = 0
        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8

        self._msh_type = MeshType(int.from_bytes(data[seeker:seeker + 8], 'little'))
        seeker += 8

        self._method = MeshingMethod(int.from_bytes(data[seeker:seeker + 8], 'little'))
        seeker += 8

        self._sizing = unpack('<d', data[seeker:seeker + 8])
        seeker += 8

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._nc = utils.deserialize_array(data[seeker:seeker + data_len])
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._inz = utils.deserialize_dict(data[seeker:seeker + data_len])
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._gp_loc = utils.deserialize_dict(data[seeker:seeker + data_len])
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._sj = utils.deserialize_dict(data[seeker:seeker + data_len])
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._srf_nodes = utils.deserialize_array(data[seeker:seeker + data_len])
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._srf_inz = utils.deserialize_dict(data[seeker:seeker + data_len])
        seeker += data_len

        geometry_id = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        if isinstance(self.geometry, geo.Geometry):
            self.geometry.obj_index = geometry_id
        else:
            self.geometry = self._server.objs[geometry_id]

    def generate(self):
        """
        Generates the mesh for the object. Returns true if sucessfull.
        """
        if not self._server.is_remote_server:
            self._refresh_stamp += 1
            return True
        self.__lazy_pull__()
        res = self.call_method('generate')
        return res

    def to_gmsh(self):
        """
        Loads the mesh into gmsh
        """
        self.__lazy_pull__()
        entity = gmsh.model.addDiscreteEntity(3)
        if len(self.inz) > 0:
            if self.msh_type == MeshType.LINEAR:
                gmsh.model.mesh.setOrder(1)
            else:
                gmsh.model.mesh.setOrder(2)
        gmsh.model.mesh.addNodes(3, entity, np.arange(1, self.nc.shape[0] + 1), self.nc.reshape(-1))
        e = 1
        elem_tags = {}
        for et, inz in self.inz.items():
            elem_tags[et] = np.arange(e, e + inz.shape[0])
            e = inz.shape[0] + 1

        for et, inz in self.inz.items():
            gmsh.model.mesh.addElements(3, entity, [et], [elem_tags[et]], [inz.reshape(-1) + 1])

    def plot(self):
        """
        Plots the mesh using gmsh
        """
        self.__lazy_pull__()
        self.to_gmsh()
        gmsh.fltk.run()

    @property
    def method(self):
        """
        The used meshing method
        """
        self.__lazy_pull__()
        return self._method

    @method.setter
    def method(self, method):
        """
        The used meshing method
        """
        self._method = method
        self.__update_remote__()

    @property
    def sizing(self):
        """
        The sizing target for the mesh
        """
        self.__lazy_pull__()
        return self._sizing

    @sizing.setter
    def sizing(self, sizing):
        """
        The underlying geometry of the mesh
        """
        self._sizing = sizing
        self.__update_remote__()

    @property
    def geometry(self):
        """
        The underlying geometry of the mesh
        """
        self.__lazy_pull__()
        return self._geometry

    @geometry.setter
    def geometry(self, geometry):
        """
        The underlying geometry of the mesh
        """
        self._geometry = geometry
        self.__update_remote__()

    @property
    def msh_type(self):
        """
        Type of the mesh
        """
        self.__lazy_pull__()
        return self._msh_type

    @msh_type.setter
    def msh_type(self, msh_type):
        """
        Type of the mesh
        """
        self._msh_type = msh_type
        self.__update_remote__()

    @property
    def nc(self):
        """
        n x 3 matrix of all node coordinates of the mesh
        """
        self.__lazy_pull__()
        return self._nc

    @nc.setter
    def nc(self, nc):
        """
        n x 3 matrix of all node coordinates of the mesh
        """
        self._nc = nc
        self.__update_remote__()

    @property
    def inz(self):
        """
        inz is a dictionary of the incidence matrices for all element types of the mesh.
        Key of the dictionaries are the type IDs.
        """
        self.__lazy_pull__()
        return self._inz

    @inz.setter
    def inz(self, inz):
        """
        n x 3 matrix of all node coordinates of the mesh
        """
        self._inz = inz
        self.__update_remote__()

    @property
    def gp_loc(self):
        """
        Dictionary of the coordinates of all gauss points for each element and each element type.
        Key of the dictionaries are the type IDs.
        """
        self.__lazy_pull__()
        return self._gp_loc

    @property
    def sj(self):
        """
        Dictionary of the scaled jacobians of all gauss points for each element and each element type.
        Key of the dictionaries are the element type IDs.
        """
        self.__lazy_pull__()
        return self._sj

    @property
    def srf_nodes(self):
        """
        A vector of all node indices which are on the surface of the mesh
        """
        self.__lazy_pull__()
        return self._srf_nodes

    @property
    def srf_inz(self):
        """
        A dictionary with the incidence matrices of all surface element types.
        Key of the dictionaries are the surface element type IDs.
        """
        return self._srf_inz


@dataclass
class ElementType:
    """holdes information for one element type"""

    def __init__(self):
        self.obj_index = 0
        self.npe = 0
        self.ngp = 0
        self.shp = np.zeros([0, 0, 0])
        self.gp_loc = np.zeros([0, 3])
        self.node_loc = np.zeros([0, 3])
        self.weights = np.zeros(0)



import multiprocessing
import pathlib
import os.path
from src.simtwin.broker import Broker


def __mock_remote_server__(port, api_key):
    ssl_path = pathlib.Path(__file__).resolve().parent
    broker = Broker(port, lambda x: x == api_key, os.path.join(ssl_path, 'selfsigned.crt'),
                    os.path.join(ssl_path, 'private.key'))
    broker.await_client()
    while True:
        if not broker.listen():
            break
    broker.socket.close()


def run_remote_server(port, key):
    process = multiprocessing.Process(target=__mock_remote_server__, args=(port, key))
    process.start()
    return process



# pylint: disable-all
"""
Quirk to guarantee correct Pythonpath and single source of truth versioning
"""
from importlib.metadata import version, PackageNotFoundError
__author__ = """Henrik Stromberg"""
__email__ = 'henrik@askdrq.com'
try:
    __version__ = version('simtwin')
except PackageNotFoundError:
    __version__ = '0.0.0'

from . import broker  # noqa
from . import geometry  # noqa
from . import mesh  # noqa
from . import server  # noqa
from . import simulation  # noqa



"""
Contains helper functions needed for internal purposes of the
simtwin library. Not intended for direct use!
"""
from struct import pack, unpack
import os
import numpy as np
import rootpath
from . import exceptions


def serialize_list(lst):
    """
    Serializes a list to a stream ob bytes.
    """
    res = len(lst).to_bytes(64, 'little')
    for item in lst:
        item_byts = obj_to_bytes(item)
        res += (str(type(item)) + '|' + str(len(item_byts))).ljust(256).encode('UTF-8')
        res += item_byts
    return res


def deserialize_list(byts):
    """
    Deserializes a stream ob bytes into a list.
    """
    res = []
    n_items = int.from_bytes(byts[:64], 'little')
    seeker = 64
    for _n in range(n_items):
        item_header = byts[seeker:seeker + 256].decode('UTF-8').strip()
        item_type = item_header.split('|')[0]
        item_len = int(item_header.split('|')[1])
        seeker += 256
        item_bytes = byts[seeker: seeker + item_len]
        res += [obj_frm_bytes(item_bytes, item_type)]
        seeker += item_len
    return res


def serialize_dict(dct):
    """
    Serializes a dictionary to a stream ob bytes.
    """
    res = len(dct).to_bytes(64, 'little')
    for key in dct:
        key_byts = obj_to_bytes(key)
        value_byts = obj_to_bytes(dct[key])
        res += (str(type(key)) + '|' + str(len(key_byts))).ljust(256).encode('UTF-8')
        res += key_byts
        res += (str(type(dct[key])) + '|' + str(len(value_byts))).ljust(256).encode('UTF-8')
        res += value_byts
    return res


def deserialize_dict(byts):
    """
    Deserializes a stream ob bytes into a dictionary.
    """
    res = {}
    n_items = int.from_bytes(byts[:64], 'little')
    seeker = 64
    for _n in range(n_items):
        key_header = byts[seeker:seeker + 256].decode('UTF-8').strip()
        key_type = key_header.split('|')[0]
        key_len = int(key_header.split('|')[1])
        seeker += 256
        key_bytes = byts[seeker: seeker + key_len]
        key = obj_frm_bytes(key_bytes, key_type)
        seeker += key_len

        value_header = byts[seeker:seeker + 256].decode('UTF-8').strip()
        value_type = value_header.split('|')[0]
        value_len = int(value_header.split('|')[1])
        seeker += 256
        value_bytes = byts[seeker: seeker + value_len]
        value = obj_frm_bytes(value_bytes, value_type)
        seeker += value_len

        res[key] = value
    return res


def get_id_list(obj):
    """
    Returns a list of IDs from a list of simtwin objects
    """
    return [item.obj_index for item in obj]


def cmp_list(lst_a, lst_b, ignore_order=True):
    """
    Returns true if all the lists contain the same items and no others.
    If ignore_order is set, the order of items is ignored for the comparison.
    """
    if len(lst_a) != len(lst_b):
        return False
    if ignore_order:
        for item in lst_a:
            if item not in lst_b:
                return False
    else:
        for i, _ in enumerate(lst_a):
            if not lst_a[i] == lst_b[i]:
                return False
    return True


def cmp_dict(dct_a, dct_b):
    """
    Compares to dictionaries. Returns true if both contain the same keys and values.
    """
    if len(dct_a) != len(dct_b):
        res = False
    else:
        res = True
    for key in dct_a:
        if key not in dct_b:
            res = False
            break
        if not dct_a[key].__class__ == dct_b[key].__class__:
            res = False
            break
        if isinstance(dct_a[key], np.ndarray):
            if not dct_a[key].shape == dct_b[key].shape:
                res = False
                break
            res = (dct_a[key] == dct_b[key]).all() or res
        else:
            res = dct_a[key] == dct_b[key]
    return res


def obj_frm_bytes(byts, type_str):
    """
    Deserializes an object to a stream of bytes.
    Throws an exception if the type of the object is not supported.
    """
    # pylint: disable=R0911
    if type_str == "<class 'int'>":
        return int.from_bytes(byts, 'little')
    if type_str == "<class 'float'>":
        return unpack('<d', byts)[0]
    if type_str == "<class 'str'>":
        return byts.decode('UTF-8')
    if type_str == "<class 'bool'>":
        return bool.from_bytes(byts, 'little')
    if type_str == "<class 'list'>":
        return deserialize_list(byts)
    if type_str == "<class 'numpy.ndarray'>":
        return deserialize_array(byts)
    if type_str == "<class 'NoneType'>":
        return None
    raise exceptions.SerializationException('Object of type ' + type_str + 'cannot be serialized.')


def obj_to_bytes(obj):
    """
    Tests the serializability of an object and returns the serialized bytes
    """
    if hasattr(obj, 'to_bytes'):
        byts = obj.to_bytes(64, 'little')
    elif hasattr(obj, '__serialize__'):
        byts = obj.__serialize__()
    elif hasattr(obj, 'encode'):
        byts = obj.encode('UTF-8')
    elif isinstance(obj, float):
        byts = pack('<d', obj)
    elif isinstance(obj, list):
        byts = serialize_list(obj)
    elif isinstance(obj, dict):
        byts = serialize_dict(obj)
    elif isinstance(obj, np.ndarray):
        byts = serialize_array(obj)
    elif obj is None:
        byts = bytes(0)
    else:
        raise exceptions.SerializationException('Object of type ' + str(type(obj)) + 'cannot be serialized.')
    return byts


def serialize_array(arr):
    """
    Serializes a numpy array
    """
    return (str(arr.dtype).ljust(8) + str(arr.shape).ljust(16)).encode() + arr.tobytes()


def deserialize_array(byts):
    """
    Forms a numpy array from serialized data
    """
    dtype = byts[:8].decode().strip()
    shape_str = byts[8:24].decode().strip()[1:-1]
    shape = []
    for segment in shape_str.split(','):
        if segment.strip() != '':
            shape += [int(segment)]
    return np.frombuffer(byts[24:], dtype=dtype).reshape(shape)


def rcv_bytes(connection, size=256):
    """
    Receives the designated number of bytes from the provided connection.
    """
    res = bytes()
    chunck_size = 1024
    for _n in range(int(size / chunck_size) + 2):
        bytes_read = connection.recv(min(size - len(res), chunck_size))
        res += bytes_read
        if len(res) == size:
            break
    return res


def find_cert_file():
    """
    Locates SSL certificate file
    """
    module_dir = rootpath.detect()
    crt_files = []
    for dirpath, _, filenames in os.walk(module_dir):
        for filename in filenames:
            if filename.endswith('.crt'):
                crt_files.append(os.path.join(dirpath, filename))
    crt_files = sorted(crt_files, key=lambda x: os.path.basename(x))
    if len(crt_files) == 0:
        exceptions.DrqException('No SSL certificate found in root dir.')
    return crt_files[0]



"""
Contains the server class
"""

from sys import maxsize
from weakref import WeakValueDictionary
from os.path import exists
import socket
import ssl
import configparser
from .__init__ import __version__  # pylint: disable=E0402
from . import utils
from . import exceptions


class Server:
    """
    An instance of the class object represents the connection to a numeric server.
    """
    # pylint: disable=R0913
    # pylint: disable=R0902
    def __init__(self, config_path='', server='', port=-1, api_key='', timeout=5., cert_path=''):
        if exists(config_path):
            self.__init_with_config__(api_key, config_path, port, server, cert_path)
        elif server == '' or port == -1 or api_key == '':
            raise exceptions.CommunicationException('Port, server or API key to connect to not specified!')
        else:
            self._address = server
            self._port = port
            self._api_key = api_key
            self._cert_path = cert_path
        if not exists(self._cert_path):
            self._cert_path = utils.find_cert_file()
        self.context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        self.context.load_verify_locations(cafile=self._cert_path)
        self.is_remote_server = True
        self._objs = WeakValueDictionary()
        self._socket = socket.socket()
        self._socket = self.context.wrap_socket(self._socket, server_hostname=self._address)
        self._socket.settimeout(timeout)
        self._socket.connect((self._address, self._port))

        self._socket.sendall(__version__.ljust(256).encode())
        self._socket.sendall(self._api_key.ljust(256).encode())
        answer = utils.rcv_bytes(self._socket).decode('UTF-8').strip()
        if answer == 'OK':
            return
        if answer == 'INVALID_API_KEY':
            raise exceptions.AuthenticationException('API key is not valid')

    def add_obj(self, obj):
        """
        Adds an object to the server and returns the obj_index it was registered with.
        """
        obj_index = self.__get_free_id__()
        self._objs[obj_index] = obj
        return obj_index

    def get_obj(self, obj_index):
        """
        Returns a reference to the server managed object with the given obj_index
        """
        if obj_index not in self._objs:
            if not self.pull(obj_index):
                raise exceptions.ServerException('Object exists neither in client nor in server')
        return self._objs[obj_index]

    def push(self, obj_index, force=False):
        """
        Pushes the object stored at obj_index to the remote server.
        """
        obj = self._objs[obj_index]
        if obj.refresh_stamp <= self.get_refresh_stamp(obj_index) and not force:
            return
        type_string = str(type(obj))
        data = obj.__serialize__()
        package = ('PUSH|' + str(obj_index) + '|' + type_string + '|' + str(obj.refresh_stamp) +
                   '|' + str(len(data))).ljust(256).encode('UTF-8') + data
        self._socket.sendall(package)
        answer = utils.rcv_bytes(self._socket).decode('UTF-8')
        if answer.strip() != 'OK':
            raise exceptions.ServerException(answer.split('|')[1])

    def get_refresh_stamp(self, obj_index):
        """
        gets the current server side refresh stamp from the server
        """
        package = ('GET_STAMP|' + str(obj_index)).ljust(256).encode('UTF-8')
        self._socket.sendall(package)
        package = utils.rcv_bytes(self._socket, 256).decode('UTF-8')
        if not int(package.split('|')[1]) == obj_index:
            raise exceptions.CommunicationException('answer with different object obj_index received')
        return int(package.split('|')[2])

    def pull(self, obj_index, force=False):
        """
        Pulls the object sored at obj_index from the server and returns true on success.
        """
        stamp_at_server = self.get_refresh_stamp(obj_index)
        if stamp_at_server <= self._objs[obj_index].refresh_stamp and not force:
            return True

        package = ('PULL|' + str(obj_index)).ljust(256).encode('UTF-8')
        self._socket.sendall(package)
        header = utils.rcv_bytes(self._socket).decode('UTF-8')
        if header[0][:15] == 'NO_OBJECT_AT_ID':
            return False
        if str(obj_index) != header.split('|')[1]:
            raise exceptions.CommunicationException('answer with different object obj_index received')
        data = utils.rcv_bytes(self._socket, int(header.split('|')[2]))
        self._objs[obj_index].__deserialize__(data)
        return True

    def remove(self, obj_index):
        """
        Removes the object sored at obj_index from the server
        """
        package = ('REMOVE|' + str(obj_index)).ljust(256).encode('UTF-8')
        self._socket.sendall(package)
        del self._objs[obj_index]

    def call_method(self, name, obj_index, args, return_type="<class 'bool'>"):
        """
        Calls a method of on object on the remote server and passes back eventual results.
        """
        arg_bytes = utils.serialize_dict(args)
        package = ('CALL|' + str(obj_index) + '|' + name + '|' + str(len(arg_bytes))).ljust(256).encode()
        self._socket.sendall(package)
        self._socket.sendall(arg_bytes)
        reply = utils.rcv_bytes(self._socket).decode('UTF-8')
        if reply.strip() != 'OK':
            raise exceptions.ServerException(reply.strip())
        reply = utils.rcv_bytes(self._socket).decode('UTF-8')
        return_len = int(reply.split('|')[1])
        return_data = utils.rcv_bytes(self._socket, return_len)
        return utils.obj_frm_bytes(return_data, return_type)

    def __get_free_id__(self):
        """
        Returns the smallest free obj_index
        """
        res = 0
        for obj_index in range(maxsize):
            if obj_index not in self._objs:
                res = obj_index
                break
        return res

    def __init_with_config__(self, api_key, config_path, port=-1, server='', cert_path=''):
        """
        Initializes a server object using a config file
        """
        config = configparser.ConfigParser()
        config.read(config_path)
        if server == '':
            self._address = config['DEFAULT']['server']
            if self._address in ('localhost', '127.0.0.1'):
                self._address = socket.gethostname()
        else:
            self._address = server
        if port == -1:
            self._port = int(config['DEFAULT']['port'])
        else:
            self._port = port
        if api_key == '':
            self._api_key = config['DEFAULT']['api_key']
        else:
            self._api_key = api_key
        if cert_path == '':
            self._cert_path = config['DEFAULT']['cert_path']
        else:
            self._cert_path = cert_path

    def __eq__(self, other):
        return self._address == other._address and self._port == other._port

    def close(self):
        """
        Orders the server to close
        """
        package = 'CLOSE'.ljust(256).encode()
        self._socket.sendall(package)

    def __del__(self):
        self.close()



# pylint: disable=C0116
# pylint: disable=C0302
"""
Contains classes to represent simulation geometry
"""

from dataclasses import dataclass
from os.path import exists
from os import remove
import tempfile
import numpy as np
import gmsh
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from . import utils, exceptions
from .simtwin_base_classes import SimtwinEnum, SimtwinObjectBase


class Geometry(SimtwinObjectBase):
    """
    Holds simulation geometry
    """
    # pylint: disable=R0902
    def __init__(self, server, obj_index=None):
        self._slds = []
        self._srfs = []
        self._crvs = []
        self._cnrs = []
        self._files = []
        self._tessellation = Tessellation(server)
        super().__init__(server, obj_index)

    def __eq__(self, other):
        if self.refresh_stamp == -1 or other.refresh_stamp == -1:
            return self.obj_index == other.obj_index
        return utils.cmp_list(self.slds, other.slds) and \
            utils.cmp_list(self.srfs, other.srfs) and \
            utils.cmp_list(self.crvs, other.crvs) and \
            utils.cmp_list(self.cnrs, other.cnrs) and \
            utils.cmp_list(self.files, other.files) and \
            self.tessellation == other.tessellation

    def __hash__(self):
        return self.obj_index

    def __serialize__(self):
        res = bytes()
        res += self.obj_index.to_bytes(8, 'little')

        ids = utils.get_id_list(self.slds)
        slds_bytes = utils.serialize_list(ids)
        res += int.to_bytes(len(slds_bytes), 8, 'little')
        res += slds_bytes

        ids = utils.get_id_list(self._srfs)
        srfs_bytes = utils.serialize_list(ids)
        res += int.to_bytes(len(srfs_bytes), 8, 'little')
        res += srfs_bytes

        ids = utils.get_id_list(self._crvs)
        crvs_bytes = utils.serialize_list(ids)
        res += int.to_bytes(len(crvs_bytes), 8, 'little')
        res += crvs_bytes

        ids = utils.get_id_list(self._cnrs)
        cnrs_bytes = utils.serialize_list(ids)
        res += int.to_bytes(len(cnrs_bytes), 8, 'little')
        res += cnrs_bytes

        ids = utils.get_id_list(self._files)
        files_bytes = utils.serialize_list(ids)
        res += int.to_bytes(len(files_bytes), 8, 'little')
        res += files_bytes
        tessellation_bytes = self._tessellation.__serialize__()
        res += int.to_bytes(len(tessellation_bytes), 8, 'little')
        res += tessellation_bytes
        return res

    def __deserialize__(self, data):
        seeker = 0
        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8

        data_bytes = data[seeker:seeker + data_len]
        seeker += data_len
        slds_list = utils.deserialize_list(data_bytes)
        self._slds = []
        for sld in slds_list:
            self._slds += [self._server.objs[sld]]

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        data_bytes = data[seeker:seeker + data_len]
        seeker += data_len
        srfs_list = utils.deserialize_list(data_bytes)
        self._srfs = []
        for srf in srfs_list:
            self._srfs += [self._server.objs[srf]]

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        data_bytes = data[seeker:seeker + data_len]
        seeker += data_len
        crvs_list = utils.deserialize_list(data_bytes)
        self._crvs = []
        for crv in crvs_list:
            self._crvs += [self._server.objs[crv]]

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        data_bytes = data[seeker:seeker + data_len]
        seeker += data_len
        crnr_list = utils.deserialize_list(data_bytes)
        self._cnrs = []
        for crnr in crnr_list:
            self._cnrs += [self._server.objs[crnr]]

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        data_bytes = data[seeker:seeker + data_len]
        seeker += data_len

        files_list = utils.deserialize_list(data_bytes)
        self._files = []
        for file_obj_index in files_list:
            self._files += [self._server.objs[file_obj_index]]

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        data_bytes = data[seeker:seeker + data_len]
        seeker += data_len
        self.tessellation.__deserialize__(data_bytes)

    @property
    def slds(self):
        """
        List of all solids in the geometry
        """
        return self._slds

    @property
    def srfs(self):
        """
        List of all surfaces in the geometry
        """
        return self._srfs

    @property
    def crvs(self):
        """
        List of all curves in the geometry
        """
        return self._crvs

    @property
    def cnrs(self):
        """
        List of all corners in the geometry
        """
        return self._cnrs

    @property
    def files(self):
        """
        List of all files composing the geometry
        """
        return self._files

    @files.setter
    def files(self, files):
        self._files = files
        self.__push__()
        self.__pull__()

    @property
    def tessellation(self):
        """
        Tesselation of the geometry for rendering
        """
        return self._tessellation


class GeometryFile(SimtwinObjectBase):
    """Holds information on a loaded file and applied transformations"""
    # pylint: disable=R0902

    def __init__(self, server, file_path='', translation=np.zeros(3), rotations=None, obj_index=None):
        # pylint: disable=R0913
        self._is_tmp_file = False
        if file_path != '':
            if not exists(file_path):
                raise FileExistsError
        self._path = file_path
        self._translation = translation
        if rotations is None:
            self.rotations = []
        elif isinstance(rotations, Rotation):
            self.rotations = [rotations]
        elif isinstance(rotations, list):
            self.rotations = rotations
        else:
            raise exceptions.DrqException('Unknown rotation type')
        self._file = None
        super().__init__(server, obj_index)

    def __eq__(self, other):
        if self.refresh_stamp == -1 or other.refresh_stamp == -1:
            return self.obj_index == other.obj_index
        return self.file_bytes == other.file_bytes and \
            (self._translation == other.translation).all() and \
            self._rotations == other.rotations

    def __hash__(self):
        return self.obj_index

    def __serialize__(self):
        res = bytes()
        res += self.obj_index.to_bytes(8, 'little')
        if self._path == '':
            file_data = bytes()
        else:
            file_data = self.file_bytes

        res += len(file_data).to_bytes(8, 'little')
        res += file_data

        data = utils.obj_to_bytes(self.translation)
        res += len(data).to_bytes(8, 'little')
        res += data

        rot_ids = utils.get_id_list(self.rotations)
        data = utils.serialize_list(rot_ids)
        res += len(data).to_bytes(8, 'little')
        res += data
        return res

    def __deserialize__(self, data):
        seeker = 0
        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._is_tmp_file = True
        file_bytes_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        file_byte = data[seeker:seeker+file_bytes_len]
        seeker += file_bytes_len
        # pylint: disable=R1732
        self._file = tempfile.NamedTemporaryFile(delete=False, suffix='.step')
        self._path = self._file.name

        with open(self.path, 'wb') as fp:
            fp.write(file_byte)

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self.translation = utils.deserialize_array(data[seeker:seeker + data_len])
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        rot_ids = utils.deserialize_list(data[seeker:seeker + data_len])
        seeker += data_len
        for obj_index in rot_ids:
            self.rotations += [self._server.objs[obj_index]]

    def __del__(self):
        if self.path != '' and self._is_tmp_file:
            remove(self.path)
        super().__del__()

    @property
    def path(self):
        """
        Path to the file
        """
        return self._path

    @property
    def file_bytes(self):
        """
        Returns the bytes of the underlying geometry file.
        """
        if self._path == '':
            return bytes()
        with open(self._path, 'rb') as fp:
            file_bytes = fp.read()
        return file_bytes

    @property
    def translation(self):
        """
        Translation applied to the geometry of the file
        """
        return self._translation

    @translation.setter
    def translation(self, translation):
        self._translation = translation

    @translation.deleter
    def translation(self):
        self._translation = np.zeros(3)

    @property
    def rotations(self):
        """
        List of rotations applied to the geometry of the file
        """
        return self._rotations

    @rotations.setter
    def rotations(self, rotations):
        self._rotations = rotations

    @rotations.deleter
    def rotations(self):
        self.rotations = []


@dataclass()
class Rotation(SimtwinObjectBase):
    """
    Represents a rotation
    cnt: any point on the rotation axis
    axis: vector of the rotation axis
    angle: rotation angle in degrees
    """

    def __init__(self, server, cnt=np.zeros(3), axis=np.zeros(3), angle=0., obj_index=None):  # pylint: disable=R0913
        self._cnt = cnt
        self._axis = axis
        self._angle = angle
        super().__init__(server, obj_index)

    @property
    def cnt(self):
        self.__pull__(force=True)
        return self._cnt

    @cnt.setter
    def cnt(self, cnt):
        self._cnt = cnt
        self.__push__()

    @property
    def axis(self):
        self.__pull__(force=True)
        return self._axis

    @axis.setter
    def axis(self, axis):
        self._axis = axis
        self.__push__()

    @property
    def angle(self):
        self.__pull__(force=True)
        return self._angle

    @angle.setter
    def angle(self, angle):
        self._angle = angle
        self.__push__(force=True)

    def __eq__(self, other):
        return (self._cnt == other.cnt).all() and (self._axis == other.axis).all() and self._angle == other.angle

    def __hash__(self):
        return self.obj_index

    def __serialize__(self):
        """
        Serializes the rotation to bytes
        """
        res = bytes()
        res += self.obj_index.to_bytes(8, 'little')
        res += utils.serialize_list([self._cnt, self._axis, np.array(self._angle)])
        return res

    def __deserialize__(self, data):
        seeker = 0
        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        data = utils.deserialize_list(data[8:])
        self._cnt = data[0]
        self._axis = data[1]
        self._angle = data[2]


class GeometrySelectionType(SimtwinEnum):
    """
    Types of geometry selection
    """
    NONE = 0
    VERTEX = 1
    CURVE = 2
    FACE = 3
    BODY = 4
    ALL = 5


class GeometrySelection(SimtwinObjectBase):
    """
    Represents a user selection of geometry
    """
    def __init__(self, server, selection_type=GeometrySelectionType.BODY, geometry_id=0, obj_index=None):
        self.selection_type = selection_type
        self.geometry_id = geometry_id
        super().__init__(server, obj_index)

    def __eq__(self, other):
        if self.refresh_stamp == -1 or other.refresh_stamp == -1:
            return self.obj_index == other.obj_index
        return self.selection_type == other.selection_type and self.geometry_id == other.geometry_id

    def __hash__(self):
        return self.obj_index

    def __serialize__(self):
        res = bytes()
        res += self.obj_index.to_bytes(8, 'little')
        res += self.selection_type.value.to_bytes(8, 'little')
        res += self.geometry_id.to_bytes(8, 'little')
        return res

    def __deserialize__(self, data):
        seeker = 0
        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self.selection_type = GeometrySelectionType(int.from_bytes(data[seeker:seeker + 8], 'little'))
        seeker += 8
        self.geometry_id = int.from_bytes(data[seeker:seeker + 8], 'little')


class Solid(SimtwinObjectBase):
    """
    Represents a geometry solid
    """
    def __init__(self, server, gmsh_id=0, obj_index=None):
        self._gmsh_id = gmsh_id
        self._srfs = []
        self._crvs = []
        self._cnrs = []
        super().__init__(server, obj_index)

    def __serialize__(self):
        byts = bytes()
        byts += self.obj_index.to_bytes(8, 'little')
        byts += self.gmsh_id.to_bytes(8, 'little')

        ids = utils.get_id_list(self.srfs)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes

        ids = utils.get_id_list(self.crvs)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes

        ids = utils.get_id_list(self.cnrs)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes
        return byts

    def __deserialize__(self, data):
        seeker = 0
        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._gmsh_id = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._srfs += [self._server.objs[obj_index]]
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._crvs += [self._server.objs[obj_index]]
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._cnrs += [self._server.objs[obj_index]]
        seeker += data_len

    def __eq__(self, other):
        if self.refresh_stamp == -1 or other.refresh_stamp == -1:
            return self.obj_index == other.obj_index

        return self.obj_index == other.obj_index and \
            utils.cmp_list(utils.get_id_list(self.srfs), utils.get_id_list(other.srfs)) and \
            utils.cmp_list(utils.get_id_list(self.crvs), utils.get_id_list(other.crvs)) and \
            utils.cmp_list(utils.get_id_list(self.cnrs), utils.get_id_list(other.cnrs))

    def __hash__(self):
        return hash(hash(type(self)) + self.obj_index)

    @property
    def gmsh_id(self):
        return self._gmsh_id

    @gmsh_id.setter
    def gmsh_id(self, gmsh_id):
        self._gmsh_id = gmsh_id
        self.__update_remote__()

    @property
    def srfs(self):
        """
        References to all surfaces the solid is a part of.
        """
        return self._srfs

    @property
    def crvs(self):
        """
        References to all curves the solid is a part of.
        """
        return self._crvs

    @property
    def cnrs(self):
        """
        References to all corner the solid is a part of.
        """
        return self._cnrs


class Surface(SimtwinObjectBase):
    """
    Represents a geometry solid
    """

    def __init__(self, server, gmsh_id=0, obj_index=None):
        self._gmsh_id = gmsh_id
        self._slds = []
        self._crvs = []
        self._cnrs = []
        super().__init__(server, obj_index)

    def __serialize__(self):
        byts = bytes()
        byts += self.obj_index.to_bytes(8, 'little')
        byts += self.gmsh_id.to_bytes(8, 'little')

        ids = utils.get_id_list(self.slds)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes

        ids = utils.get_id_list(self.crvs)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes

        ids = utils.get_id_list(self.cnrs)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes
        return byts

    def __deserialize__(self, data):
        seeker = 0
        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._gmsh_id = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._slds += [self._server.objs[obj_index]]
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._crvs += [self._server.objs[obj_index]]
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._cnrs += [self._server.objs[obj_index]]
        seeker += data_len

    def __eq__(self, other):
        if self.refresh_stamp == -1 or other.refresh_stamp == -1:
            return self.obj_index == other.obj_index
        return self.obj_index == other.obj_index and \
            utils.cmp_list(utils.get_id_list(self.slds), utils.get_id_list(other.slds)) and \
            utils.cmp_list(utils.get_id_list(self.crvs), utils.get_id_list(other.crvs)) and \
            utils.cmp_list(utils.get_id_list(self.cnrs), utils.get_id_list(other.cnrs))

    def __hash__(self):
        return hash(hash(type(self)) + self.obj_index)

    @property
    def gmsh_id(self):
        return self._gmsh_id

    @gmsh_id.setter
    def gmsh_id(self, gmsh_id):
        self._gmsh_id = gmsh_id
        self.__update_remote__()

    @property
    def slds(self):
        """
        References to all solids the surface is a part of.
        """
        return self._slds

    @property
    def crvs(self):
        """
        References to all curves the surface is a part of.
        """
        return self._crvs

    @property
    def cnrs(self):
        """
        References to all corner the surface is a part of.
        """
        return self._cnrs


class Curve(SimtwinObjectBase):
    """
    Represents a geometry solid
    """

    def __init__(self, server, gmsh_id=0, obj_index=None):
        self._gmsh_id = gmsh_id
        self._slds = []
        self._srfs = []
        self._cnrs = []
        super().__init__(server, obj_index)

    def __serialize__(self):
        byts = bytes()
        byts += self.obj_index.to_bytes(8, 'little')
        byts += self.gmsh_id.to_bytes(8, 'little')

        ids = utils.get_id_list(self.slds)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes

        ids = utils.get_id_list(self.srfs)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes

        ids = utils.get_id_list(self.cnrs)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes
        return byts

    def __deserialize__(self, data):
        seeker = 0
        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._gmsh_id = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._slds += [self._server.objs[obj_index]]

        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._srfs += [self._server.objs[obj_index]]

        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._cnrs += [self._server.objs[obj_index]]

        seeker += data_len

    def __eq__(self, other):
        if self.refresh_stamp == -1 or other.refresh_stamp == -1:
            return self.obj_index == other.obj_index
        return self.obj_index == other.obj_index and \
            utils.cmp_list(utils.get_id_list(self.slds), utils.get_id_list(other.slds)) and \
            utils.cmp_list(utils.get_id_list(self.srfs), utils.get_id_list(other.srfs)) and \
            utils.cmp_list(utils.get_id_list(self.cnrs), utils.get_id_list(other.cnrs))

    def __hash__(self):
        return hash(hash(type(self)) + self.obj_index)

    @property
    def gmsh_id(self):
        return self._gmsh_id

    @gmsh_id.setter
    def gmsh_id(self, gmsh_id):
        self._gmsh_id = gmsh_id
        self.__update_remote__()

    @property
    def slds(self):
        """
        References to all solids the curve is a part of.
        """
        return self._slds

    @property
    def srfs(self):
        """
        References to all surfaces the curve is a part of.
        """
        return self._srfs

    @property
    def cnrs(self):
        """
        References to all corner the curve is a part of.
        """
        return self._cnrs


class Corner(SimtwinObjectBase):
    """
    Represents a geometry solid
    """

    def __init__(self, server, gmsh_id=0, obj_index=None):
        self._gmsh_id = gmsh_id
        self._slds = []
        self._crvs = []
        self._srfs = []
        super().__init__(server, obj_index)

    def __serialize__(self):
        byts = bytes()
        byts += self.obj_index.to_bytes(8, 'little')
        byts += self.gmsh_id.to_bytes(8, 'little')

        ids = utils.get_id_list(self.slds)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes

        ids = utils.get_id_list(self.crvs)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes

        ids = utils.get_id_list(self.srfs)
        ids_bytes = utils.serialize_list(ids)
        byts += len(ids_bytes).to_bytes(8, 'little')
        byts += ids_bytes
        return byts

    def __deserialize__(self, data):
        seeker = 0
        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._gmsh_id = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._slds += [self._server.objs[obj_index]]
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._crvs += [self._server.objs[obj_index]]
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        lst = utils.deserialize_list(data[seeker:seeker + data_len])
        for obj_index in lst:
            self._srfs += [self._server.objs[obj_index]]

        seeker += data_len

    def __eq__(self, other):
        if self.refresh_stamp == -1 or other.refresh_stamp == -1:
            return self.obj_index == other.obj_index
        return self.obj_index == other.obj_index and \
            utils.cmp_list(utils.get_id_list(self.slds), utils.get_id_list(other.slds)) and \
            utils.cmp_list(utils.get_id_list(self.crvs), utils.get_id_list(other.crvs)) and \
            utils.cmp_list(utils.get_id_list(self.srfs), utils.get_id_list(other.srfs))

    def __hash__(self):
        return hash(hash(type(self)) + self.obj_index)

    @property
    def gmsh_id(self):
        return self._gmsh_id

    @gmsh_id.setter
    def gmsh_id(self, gmsh_id):
        self._gmsh_id = gmsh_id
        self.__update_remote__()

    @property
    def slds(self):
        """
        References to all solids the corner is a part of.
        """
        return self._slds

    @property
    def crvs(self):
        """
        References to all curves the corner is a part of.
        """
        return self._crvs

    @property
    def srfs(self):
        """
        References to all surfaces the corner is a part of.
        """
        return self._srfs


class Tessellation(SimtwinObjectBase):
    """
    Contains information required to render geometry.
    """
    # pylint: disable=R0902

    def __init__(self, server, obj_index=None):
        self._nc = np.zeros([0, 3])  # Coordinates of all nodes
        self._inz = np.zeros([0, 3], dtype=int)
        self._ed = np.zeros([0, 2], dtype=int)
        self._srfs = np.zeros(0, dtype=int)
        self._slds = np.zeros(0, dtype=int)
        self._crvs = np.zeros(0, dtype=int)
        self._cnrs = np.zeros(0, dtype=int)
        super().__init__(server, obj_index)

    def __eq__(self, other):
        if self.refresh_stamp == -1 or other.refresh_stamp == -1:
            return self.obj_index == other.obj_index
        self.__lazy_pull__()
        if not other.up_to_date:
            other.__pull__()
        if self.nc.shape[0] == 0 and self.inz.shape[0] == 0 and \
                other.nc.shape[0] == 0 and other.inz.shape[0] == 0:
            return True
        if not self.nc.shape == other.nc.shape and self.inz.shape == other.inz.shape:
            return False
        if not (self.nc == other.nc).all() and (self.inz == other.inz).all():
            return False
        return (self.srfs == other.srfs).all() and \
            (self.slds == other.slds).all() and \
            (self.crvs == other.crvs).all() and \
            (self.cnrs == other.cnrs).all()

    def __hash__(self):
        self.__lazy_pull__()
        return self.obj_index

    def __serialize__(self):
        res = bytes()
        res += self.obj_index.to_bytes(8, 'little')

        nc_bytes = utils.obj_to_bytes(self._nc)
        res += int.to_bytes(len(nc_bytes), 8, 'little')
        res += nc_bytes

        inz_bytes = utils.obj_to_bytes(self._inz)
        res += int.to_bytes(len(inz_bytes), 8, 'little')
        res += inz_bytes

        ed_bytes = utils.obj_to_bytes(self._ed)
        res += int.to_bytes(len(ed_bytes), 8, 'little')
        res += ed_bytes

        faces_bytes = utils.obj_to_bytes(self._srfs)
        res += int.to_bytes(len(faces_bytes), 8, 'little')
        res += faces_bytes

        solids_bytes = utils.obj_to_bytes(self._slds)
        res += int.to_bytes(len(solids_bytes), 8, 'little')
        res += solids_bytes

        corners_bytes = utils.obj_to_bytes(self._crvs)
        res += int.to_bytes(len(corners_bytes), 8, 'little')
        res += corners_bytes

        corners_bytes = utils.obj_to_bytes(self._cnrs)
        res += int.to_bytes(len(corners_bytes), 8, 'little')
        res += corners_bytes
        return res

    def __deserialize__(self, data):
        seeker = 0

        self._obj_index = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._nc = utils.obj_frm_bytes(data[seeker:seeker + data_len], "<class 'numpy.ndarray'>")
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._inz = utils.obj_frm_bytes(data[seeker:seeker + data_len], "<class 'numpy.ndarray'>")
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._ed = utils.obj_frm_bytes(data[seeker:seeker + data_len], "<class 'numpy.ndarray'>")
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._srfs = utils.obj_frm_bytes(data[seeker:seeker + data_len], "<class 'numpy.ndarray'>")
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._slds = utils.obj_frm_bytes(data[seeker:seeker + data_len], "<class 'numpy.ndarray'>")
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._crvs = utils.obj_frm_bytes(data[seeker:seeker + data_len], "<class 'numpy.ndarray'>")
        seeker += data_len

        data_len = int.from_bytes(data[seeker:seeker + 8], 'little')
        seeker += 8
        self._cnrs = utils.obj_frm_bytes(data[seeker:seeker + data_len], "<class 'numpy.ndarray'>")
        seeker += data_len

    @property
    def nc(self):
        """
        n x 3 matrix of the coordinates of all nodes
        """
        self.__lazy_pull__()
        return self._nc

    @property
    def inz(self):
        """
        n x 3 matrix of the nodes of each triangle
        """
        self.__lazy_pull__()
        return self._inz

    @property
    def ed(self):
        """
        n x 2 matrix of all edges of the tesselation
        """
        self.__lazy_pull__()
        return self._ed

    @property
    def srfs(self):
        """
        Numpy array of the face indicis for all triangles
        """
        self.__lazy_pull__()
        return self._srfs

    @property
    def slds(self):
        """
        Numpy array of the solid indicis for all triangles
        """
        self.__lazy_pull__()
        return self._slds

    @property
    def crvs(self):
        """
        Numpy array with the curve indicis for all edges
        Edges which are not on a curve are denoted with -1
        """
        self.__lazy_pull__()
        return self._crvs

    @property
    def cnrs(self):
        """
        Numpy array with the corner indicis for all nodes
        Nodes which are not on a curve are denoted with -1
        """
        self.__lazy_pull__()
        return self._cnrs

    def plot_fltk(self):
        """
        Plots the tesselation using the fltk renderer of gmsh
        """
        self.__lazy_pull__()
        gmsh.model.add('plot')
        entity = {0: gmsh.model.addDiscreteEntity(0), 1: gmsh.model.addDiscreteEntity(1),
                  2: gmsh.model.addDiscreteEntity(2), 3: gmsh.model.addDiscreteEntity(3)}
        gmsh.model.mesh.addNodes(3, entity[3], np.arange(1, self.nc.shape[0] + 1), self.nc.reshape(-1))
        gmsh.model.mesh.addElements(2, entity[2], [2], [np.arange(self.inz.shape[0]) + 1], [self.inz.flatten() + 1])
        gmsh.fltk.run()
        gmsh.model.remove()

    def plot_matplotlib(self, ax=None, color=np.array([0, 0, 0])):
        """
        or matplotlib.
        Plots the tesselation using matplotlib. An axis to plot into and plotting color can be specified.
        Returns the axis which was provided or created.
        """
        self.__lazy_pull__()
        if isinstance(color, list):
            color = np.array(color)
        if isinstance(color, int):
            color = float(color)
        elif isinstance(color, float):
            color = np.tile(color, (self.ed.shape[0], 3))
        elif len(color.shape) == 1:
            if color.shape[0] == 3:
                color = np.tile(color.T, (self.ed.shape[0], 1))
            elif color.shape[0] == 1:
                color = np.tile(color.T, (self.ed.shape[0], 3))
            else:
                color = np.stack([color, 1 - color, np.zeros_like(color)]).T
        if ax is None:
            fig = plt.figure()
            ax = Axes3D(fig, auto_add_to_figure=False)
            fig.add_axes(ax)
            ax.set_xlabel('X')
            ax.set_ylabel('Y')
            ax.set_zlabel('Z')

        for i in range(self.ed.shape[0]):
            ax.plot(self.nc[self.ed[i], 0], self.nc[self.ed[i], 1], self.nc[self.ed[i], 2], color=color[i])
        return ax



"""
Custom exceptions for the simtwin library
"""


class CommunicationException(Exception):
    """
    General exception in client and ser communication.
    """
    def __init__(self, message):
        self.message = message


class AuthenticationException(Exception):
    """
    The authentication of the client against the server is rejected
    """
    def __init__(self, message):
        self.message = message


class ServerException(Exception):
    """
    Internal server error
    """
    def __init__(self, message):
        self.message = message


class SerializationException(Exception):
    """
    Failure to serialize object
    """
    def __init__(self, message):
        self.message = message


class DrqException(Exception):
    """
    General exception in the client side library
    """
    def __init__(self, message):
        self.message = message



"""
The object broker manages all objects created through the API on the remote server.
"""

import socket
import signal
from sys import maxsize
import ssl
from . import utils
from . import geometry
from . import mesh
from . import simulation
from . import exceptions

object_types = {'Geometry': geometry.Geometry,
                'GeometryFile': geometry.GeometryFile,
                'Rotation': geometry.Rotation,
                'GeometrySelection': geometry.GeometrySelection,
                'Solid': geometry.Solid,
                'Surface': geometry.Surface,
                'Curve': geometry.Curve,
                'Corner': geometry.Corner,
                'Tessellation': geometry.Tessellation,
                'Mesh': mesh.Mesh,
                'Simulation': simulation.Simulation,
                'CoordinateSystem': simulation.CoordinateSystem,
                'BC': simulation.BC,
                'Question': simulation.Question,
                'Solution': simulation.Solution}


class Broker:
    """
    The broker class manages the created objects within the remote server
    """
    # pylint: disable=R0902
    def __init__(self, port=1441, key_validator=lambda x: True, cert_path='', key_path=''):
        signal.signal(signal.SIGTERM, self.__handle_signal__)
        self.is_remote_server = False
        self.objs = {}
        self.context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        self.context.load_cert_chain(certfile=cert_path, keyfile=key_path)
        self.socket = socket.socket()
        self.socket.bind(('', port))
        self.socket.listen(5)
        self.key_validator = key_validator
        self.api_key = ''
        self.connection = None
        self.addr = None
        self.drq_version = None
        self.connected = False

    def await_client(self):
        """
        Lets the broker wait for a client to login
        """
        self.connection, self.addr = self.socket.accept()  # Establish connection with client.
        self.connection = self.context.wrap_socket(self.connection, server_side=True)
        self.drq_version = utils.rcv_bytes(self.connection).decode().strip()
        self.api_key = utils.rcv_bytes(self.connection).decode().strip()
        if self.key_validator(self.api_key):
            self.connection.sendall('OK'.ljust(256).encode())
            self.connected = True
        else:
            self.connection.sendall('INVALID_API_KEY'.ljust(256).encode())
            self.api_key = ''

    def __handle_signal__(self):
        if self.socket:
            self.socket.close()

    def add_obj(self, obj, obj_index=-1):
        """
        Adds the provided object to the broker with the given obj_index
        """
        if obj_index == -1:
            obj_index = self.__get_free_id__()
        self.objs[obj_index] = obj
        return obj_index

    def __get_free_id__(self):
        for res in range(maxsize):
            if res not in self.objs:
                return res
        raise exceptions.ServerException('Too many objects in server')

    def listen(self):
        """
        Listen to the remote client for further commands
        """
        keep_server = True
        if not self.connected:
            return keep_server
        rbuff = utils.rcv_bytes(self.connection)
        command = rbuff.decode('UTF-8')

        if __is_cmd__(command, 'PUSH'):
            self.__cmd_push__(command)
        elif __is_cmd__(command, 'PULL'):
            self.__cmd_pull__(command)
        elif __is_cmd__(command, 'REMOVE'):
            self.__cmd_remove__(command)
        elif __is_cmd__(command, 'GET_STAMP'):
            self.__cmd_get_stamp__(command)

        elif __is_cmd__(command, 'CALL'):
            self.__cmd_call__(command)
            keep_server = True
        elif __is_cmd__(command, 'CLOSE'):
            self.__cmd_close__()
        return keep_server

    def __cmd_close__(self):
        self.connection.close()
        self.objs = {}

    def __cmd_call__(self, command):
        obj_index = int(command.split('|')[1])
        method_name = command.split('|')[2]
        data_len = int(command.split('|')[3])
        args_data = utils.rcv_bytes(self.connection, data_len)
        args_dict = utils.deserialize_dict(args_data)
        if method_name in self.objs[obj_index].methods:
            self.connection.sendall('OK'.ljust(256).encode())
        else:
            self.connection.sendall('NO_SUCH_METHOD_IN_OBJECT'.ljust(256).encode())
        res = self.objs[obj_index].call_method(method_name, args_dict)
        res_data = utils.obj_to_bytes(res)
        self.connection.sendall(('RETURN_VALUE|' + str(len(res_data)) + '|' + str(type(res))).ljust(256).encode())
        self.connection.sendall(res_data)

    def __cmd_get_stamp__(self, command):
        obj_index = int(command.split('|')[1])
        if obj_index in self.objs:
            refresh_stamp = self.objs[obj_index].refresh_stamp
        else:
            refresh_stamp = -1
        package = ('STAMP|' + str(obj_index) + '|' + str(refresh_stamp)).ljust(256).encode('UTF-8')
        self.connection.sendall(package)

    def __cmd_remove__(self, command):
        obj_index = int(command.split('|')[1])
        del self.objs[obj_index]

    def __cmd_pull__(self, command):
        obj_index = int(command.split('|')[1])
        data = self.objs[obj_index].__serialize__()
        package = ('DATA|' + str(obj_index) + '|' + str(len(data))).ljust(256).encode('UTF-8') + data
        self.connection.sendall(package)

    def __cmd_push__(self, command):
        obj_index = int(command.split('|')[1])
        obj_type = command.split('|')[2]
        obj_type = obj_type.split('.')[-1].replace('>', '').replace("'", '')
        refresh_stamp = int(command.split('|')[3])
        data_len = int(command.split('|')[4])
        data = utils.rcv_bytes(self.connection, data_len)
        type_from_client = object_types[obj_type]
        if obj_index not in self.objs:
            self.objs[obj_index] = type_from_client(self, obj_index=obj_index)
        if not isinstance(self.objs[obj_index], type_from_client):
            package = ('ERROR| This index is already in use by another object of type ' +
                       str(type(self.objs[obj_index]))).ljust(256).encode('UTF-8')
            self.connection.sendall(package)
        else:
            self.objs[obj_index].__deserialize__(data)
            self.objs[obj_index]._refresh_stamp = refresh_stamp  # pylint: disable=W0212
            self.connection.sendall('OK'.ljust(256).encode())

    def push(self, obj_index, force=False):
        """
        This dummy method is required to stop a communication cycle between client and host
        """

    def pull(self, obj_index, force=False):
        """
        This dummy method is required to stop a communication cycle between client and host
        """

    def remove(self, obj_index):
        """
        This dummy method is required to stop a communication cycle between client and host
        """

    def __del__(self):
        self.socket.close()


def __is_cmd__(string, cmd):
    if len(string) < len(cmd):
        return False
    return string[:len(cmd)] == cmd
